# MIME type aliases for semantic convenience
TYPES = {
    "text": "text/plain",
    "html": "text/html",
    "json": "application/json",
    "xml": "application/xml",
    "urlencoded": "application/x-www-form-urlencoded",
    "form": "application/x-www-form-urlencoded",
    "form-data": "application/x-www-form-urlencoded",
}
